import logging
from datetime import date, datetime
from decimal import Decimal
from functools import partial
from typing import Generator, Union

from moula.metrics import positions
from moula.types import Transaction
from moula.utils import ACCURACY
from woob.capabilities import NotAvailable
from woob.capabilities.bank import CapBank
from woob.core import Woob
from woob.exceptions import BrowserUnavailable

logger = logging.getLogger(__name__)
INV_ACC_TYPES = {"LIFE_INSURANCE", "PEA", "MARKET", "PERP"}


def get_date(dt: Union[date, datetime]) -> date:
    if isinstance(dt, date) and not isinstance(dt, datetime):
        return dt
    return dt.date()


def get_type_name(woob_item) -> str:
    return {
        getattr(woob_item, key): key[5:]
        for key in dir(woob_item)
        if key.startswith("TYPE_")
    }.get(woob_item.type) or ""


def to_transaction(account_id, woob_transaction) -> Transaction:
    remote_type = get_type_name(woob_transaction).lower()
    tr_source_id = tr_destination_id = None
    tags = []
    if woob_transaction.amount < Decimal("0"):
        tr_source_id = account_id
    else:
        tr_destination_id = account_id
    if remote_type != "transfer":
        tags.append(remote_type)

    transaction = Transaction(
        description=woob_transaction.label,
        date=get_date(woob_transaction.date),
        payment_date=get_date(woob_transaction.rdate),
        interest_date=get_date(woob_transaction.vdate),
        currency_code="EUR",
        amount=woob_transaction.amount,
        tags=tags,
        remote_type=remote_type,
        destination_id=tr_destination_id,
        source_id=tr_source_id,
    )
    if woob_transaction.id:
        transaction.external_id = woob_transaction.id
    return transaction


def get_account_type(account) -> str:
    return {
        getattr(account, key): key[5:]
        for key in dir(account)
        if key.startswith("TYPE_")
    }.get(account.type) or ""


def get_line_position(backend, account_type, account):
    def wrapped(line_name, line_id, value_type, value):
        args = (
            backend.name.upper() if backend.name else "",
            account_type,
            account.label or "",
            account.id or "",
            line_name if line_name != NotAvailable else "",
            line_id if line_id != NotAvailable else "",
            value_type,
        )
        logger.debug("seeing %s %s %s %s: %s, %s (%s) => %s", *args, value)
        return positions.labels(*args).set(value)

    return wrapped


def investments_metrics(backend, account, line_position):
    try:
        for inv in backend.iter_investment(account):
            inv_position = partial(line_position, inv.label, inv.code)
            if not inv.valuation:
                continue
            inv_position("valuation", inv.valuation)
            if inv.quantity:
                inv_position("quantity", inv.quantity)
            if inv.unitvalue:
                inv_position("par-value", inv.unitvalue)
            if not inv.diff_ratio:
                continue
            gain = inv.valuation - (inv.valuation / (inv.diff_ratio + 1))
            gain = gain.quantize(ACCURACY)
            inv_position("gain-percent", inv.diff_ratio * 100)
            inv_position("gain", gain)
    except NotImplementedError:
        logger.debug(
            "%r:%r does not support itering on accounts", backend, account
        )
    except Exception:
        logger.error(
            "%r:%r error while fetching investments", backend, account
        )


def get_accounts():
    for backend in Woob().load_backends(CapBank).values():
        try:
            for account in backend.iter_accounts():
                acc_type = get_account_type(account)
                line_position = get_line_position(backend, acc_type, account)
                line_position(
                    "balance", "balance", "valuation", account.balance
                )
                yield backend, account
                if acc_type not in INV_ACC_TYPES:
                    continue
                investments_metrics(backend, account, line_position)
        except (BrowserUnavailable, NotImplementedError):
            logger.debug("%r does not support itering on accounts", backend)
        except Exception:
            logger.exception("%r error while itering on accounts", backend)


def get_transactions(
    backend,
    account,
    account_name: str,
    account_id: str,
) -> Generator[Transaction, None, None]:
    logger.debug("getting transaction for %s:%s", backend.name, account_name)
    total = 0
    try:
        for woob_transaction in backend.iter_history(account):
            total += 1
            yield to_transaction(account_id, woob_transaction)
    except NotImplementedError:
        return
    logger.info(
        "got %d transactions for %s:%s", total, backend.name, account_name
    )



#!/usr/bin/env python
import logging
import time
from collections import defaultdict
from datetime import date, timedelta
from itertools import product

import daemon_metrics
from prometheus_client import start_http_server

from moula import firefly, woob
from moula.bootstrap import conf
from moula.merger import merge

logger = logging.getLogger("moula")
try:
    logger.setLevel(getattr(logging, conf.logging.level.upper()))
    logger.addHandler(logging.StreamHandler())
except AttributeError as error:
    raise AttributeError(
        f"{conf.logging.level} isn't accepted, only DEBUG, INFO, WARNING, "
        "ERROR and FATAL are accepted"
    ) from error


def iter_accounts():
    facc_ids = firefly.browse_accounts()
    for backend, account in woob.get_accounts():
        if account.id not in facc_ids:
            continue
        yield facc_ids[account.id], backend, account


def get_all_transactions():
    transactions = defaultdict(list)
    final = []
    for firefly_account, woob_backend, account in iter_accounts():
        for transaction in woob.get_transactions(
            woob_backend,
            account,
            firefly_account["attributes"]["name"],
            firefly_account["id"],
        ):
            transactions[firefly_account["id"]].append(transaction)
        daemon_metrics.item_result(
            conf.name, bool(transactions[firefly_account["id"]]), account.id
        )
    for date_key in "date", "interest_date":
        pairs = {tuple(sorted(p)) for p in product(transactions, repeat=2)}
        for account1, account2 in sorted(pairs):
            if account1 == account2:
                continue
            logger.debug("processing account %r and %r", account1, account2)
            for transaction in merge(
                transactions[account1],
                transactions[account2],
                delay=timedelta(days=conf.merge_date_tolerance),
                comparison_key=date_key,
            ):
                final.append(transaction)
    for tr_list in transactions.values():
        for transaction in tr_list:
            if not transaction.already_merged:
                final.append(transaction)
    return sorted(final, key=lambda transaction: transaction.date)


def main():
    account_count = 0
    for _, _, account in iter_accounts():
        account_count += 1
        daemon_metrics.item_result(conf.name, True, account.id)
    info = {
        "loop-perid": conf.loop.interval,
        "loop-lookback": conf.firefly.lookback,
        "item-count": account_count,
    }
    daemon_metrics.init(conf.name, info)

    while True:
        loop_context = daemon_metrics.LoopContext(conf.name)
        with loop_context:
            if conf.firefly.lookback:
                logger.info(
                    "will only consider data for the last %d days",
                    conf.firefly.lookback,
                )
            transactions = get_all_transactions()
            if conf.firefly.lookback:
                limit = date.today() - timedelta(days=conf.firefly.lookback)
                transactions = [
                    transaction
                    for transaction in transactions
                    if transaction.date >= limit
                ]
            firefly.push(transactions, dry_run=not conf.firefly.push)

        if (interval := conf.loop.interval - loop_context.exec_interval) > 0:
            time.sleep(interval)

        if not conf.loop.enabled:
            break


if __name__ == "__main__":
    logger.info("Starting moula")
    start_http_server(conf.prometheus.port)
    main()



from the_conf import TheConf

metaconf = {
    "config_file_environ": ["CONFIG"],
    "config_files": ["~/.config/moula/moula.yml", "/etc/moula/moula.yml"],
    "parameters": [
        {"name": {"default": "moula"}},
        {
            "firefly": [
                {
                    "default_cash_account": {
                        "default": "",
                    }
                },
                {"url": {"required": True, "type": str}},
                {
                    "lookback": {
                        "default": 0,
                        "type": int,
                        "help_txt": "number of days to look into "
                        "when comparing with firefly data",
                    }
                },
                {"token": {"required": True}},
                {"push": {"required": True, "type": bool}},
            ],
        },
        {"merge_date_tolerance": {"type": int, "default": 1}},
        {"accuracy": {"type": int, "default": 2}},
        {
            "loop": [
                {"enabled": {"type": bool, "default": False}},
                {"interval": {"type": int, "default": 240}},
            ]
        },
        {
            "prometheus": [
                {"port": {"type": int, "default": 9100}},
                {"namespace": {"default": ""}},
            ]
        },
        {"logging": [{"level": {"default": "WARNING"}}]},
    ],
}

conf = TheConf(metaconf)



import logging
from datetime import timedelta
from typing import Generator, List

from moula.types import Transaction

logger = logging.getLogger(__name__)


def _merge_two(transaction_a: Transaction, transaction_b: Transaction):
    transaction_a.source_id = (
        transaction_a.source_id or transaction_b.source_id
    )
    transaction_a.destination_id = (
        transaction_a.destination_id or transaction_b.destination_id
    )
    for ext_id in transaction_a.external_ids.union(transaction_b.external_ids):
        transaction_a.external_id = ext_id
        transaction_b.external_id = ext_id
    transaction_a.already_merged = transaction_b.already_merged = True
    return transaction_a


def _iter_merged(
    mergeable_a: List[Transaction],
    mergeable_b: List[Transaction],
    delay: timedelta,
    comparison_key: str,
) -> Generator[Transaction, None, None]:
    for tr_a in mergeable_a:
        candidates = []
        for tr_b in mergeable_b:
            if not tr_b.is_mergeable or tr_a.amount != tr_b.amount * -1:
                continue
            lower = getattr(tr_b, comparison_key) - delay
            upper = getattr(tr_b, comparison_key) + delay
            if lower <= getattr(tr_a, comparison_key) <= upper:
                candidates.append(tr_b)
        if len(candidates) == 1:
            yield _merge_two(tr_a, candidates[0])
        elif len(candidates) > 1:
            candidates2 = [
                tr_b
                for tr_b in candidates
                if tr_b.source_id == tr_a.source_id
                or tr_b.destination_id == tr_a.destination_id
            ]
            if len(candidates2) == 1:
                yield _merge_two(tr_a, candidates[0])
            elif candidates2:
                logger.error(
                    "too many candidates %r (for %r)", candidates2, candidates
                )


def merge(
    transactions_a: List[Transaction],
    transactions_b: List[Transaction],
    delay: timedelta = timedelta(days=0),
    comparison_key: str = "date",
) -> Generator[Transaction, None, None]:
    mergeable_a: List[Transaction] = []
    mergeable_b: List[Transaction] = []

    # purging
    for iterator, buffer in [
        (transactions_a, mergeable_a),
        (transactions_b, mergeable_b),
    ]:
        for transaction in iterator:
            if transaction.is_mergeable:
                buffer.append(transaction)
    yield from _iter_merged(mergeable_a, mergeable_b, delay, comparison_key)

    merged = 0
    for tr_b in mergeable_b:
        if not tr_b.is_mergeable:
            merged += 1
    logger.debug(
        "merging list a(%d) and b(%d) with %d and %d mergeable result"
        "ing in %d merged",
        len(transactions_a),
        len(transactions_b),
        len(mergeable_a),
        len(mergeable_b),
        merged,
    )



from prometheus_client import Gauge
from moula.bootstrap import conf

positions = Gauge(
    "financial_positions",
    "",
    [
        "bank",
        "account_type",
        "account_name",
        "account_id",
        "line_name",
        "line_id",
        "value_type",
    ],
    namespace=conf.prometheus.namespace,
)



import logging
from collections import defaultdict
from typing import Any, Dict, Generator, List, Optional, Tuple

import requests

from moula import indexes
from moula.bootstrap import conf
from moula.types import (
    TRANSACTIONS,
    get_default_cash_account,
    is_liability,
    set_default_cash_account,
)
from moula.utils import ZERO

logger = logging.getLogger(__name__)
FireflyTransaction = Dict
FireflyTransactions = List[FireflyTransaction]
Command = Tuple[str, Optional[str], Optional[Dict]]
Commands = Generator[Command, None, None]


def query(
    method: str,
    endpoint: str,
    params: Optional[Dict[str, str]] = None,
    body: Optional[Dict[str, Any]] = None,
):
    headers = {
        "Accept": "application/json",
        "Authorization": f"Bearer {conf.firefly.token}",
    }
    response = requests.request(
        method=method,
        headers=headers,
        url=f"{conf.firefly.url}/api/v1/{endpoint.lstrip('/')}",
        params=params,
        json=body,
        timeout=60,
    )
    response.raise_for_status()
    return response.json()


def paginate(
    method: str,
    endpoint: str,
    params: Optional[Dict[str, str]] = None,
    body: Optional[Dict[str, Any]] = None,
):
    params = params or {}
    page = 1
    one_more_page = True
    while one_more_page:
        result = query(method, endpoint, params, body)
        yield from result["data"]
        page += 1
        pagination = result["meta"]["pagination"]
        params["page"] = str(page)
        one_more_page = pagination["current_page"] < pagination["total_pages"]
        if pagination["current_page"] % 20 == 0:
            logger.debug(
                "browsed %d/%d pages",
                pagination["current_page"],
                pagination["total_pages"],
            )


def browse_accounts():
    logger.debug("fetching firefly accounts")
    facc_ids = {}
    for firefly_acc in paginate("get", "accounts"):
        if firefly_acc["attributes"]["account_number"]:
            facc_ids[firefly_acc["attributes"]["account_number"]] = firefly_acc
        if firefly_acc["attributes"]["type"] == "cash":
            set_default_cash_account(firefly_acc)
        is_liability(account=firefly_acc)
    logger.info("got %d accounts from firefly", len(facc_ids))
    return facc_ids


def __fix_title(body, transactions, candidate, existings, transaction_ids):
    cand_desc = candidate["description"]
    if len(transaction_ids) > 1 and existings[0]["attributes"].get(
        "group_title"
    ):
        existing_desc = existings[0]["attributes"]["group_title"]
    else:
        existing_desc = existings[0]["attributes"]["transactions"][0][
            "description"
        ]
    if existing_desc != cand_desc and len(cand_desc) > len(existing_desc):
        if len(transaction_ids) > 1:
            body["group_title"] = cand_desc
        elif transactions:
            transactions[0]["description"] = cand_desc
        else:
            transactions.append(
                {
                    "transaction_journal_id": transaction_ids[0],
                    "description": cand_desc,
                }
            )


def _manage_existing(
    candidates: TRANSACTIONS, existings: FireflyTransactions
) -> Commands:
    cash_account = get_default_cash_account()
    if len(candidates) == 1:
        candidate = candidates[0].payload
        body: Dict = {}
        transactions = []
        transaction_ids = [
            transaction["transaction_journal_id"]
            for transaction in existings[0]["attributes"]["transactions"]
        ]
        for transaction_id in transaction_ids:
            ffly = [
                transaction
                for transaction in existings[0]["attributes"]["transactions"]
                if transaction["transaction_journal_id"] == transaction_id
            ][0]

            for key in "source_id", "destination_id":
                no_value = not ffly.get(key) or ffly[key] == "0"
                if no_value:
                    transactions.append(
                        {
                            key: candidate[key],
                            "transaction_journal_id": transaction_id,
                            "type": candidate["type"],
                        }
                    )
                elif (
                    cash_account
                    and ffly[key] == cash_account
                    and candidate[key] != cash_account
                ):
                    transactions.append(
                        {
                            key: candidate[key],
                            "transaction_journal_id": transaction_id,
                            "type": candidate["type"],
                        }
                    )
        __fix_title(body, transactions, candidate, existings, transaction_ids)
        if body or transactions:
            if transactions:
                body["transactions"] = transactions
            yield "update", existings[0]["id"], body
        else:
            yield ".", None, None
    else:
        yield "x", None, None


def iter_operations(
    existings: FireflyTransactions, candidates: TRANSACTIONS
) -> Commands:
    if not existings:
        # no existing, we must create them all
        for candidate in candidates:
            payload = candidate.payload
            # manage firefly quirk, can't have transfer with liability account
            if payload["type"] == "transfer":
                if is_liability(payload["source_id"]):
                    payload["type"] = "deposit"
                elif is_liability(payload["destination_id"]):
                    payload["type"] = "withdrawal"
            yield "create", None, {"transactions": [payload]}
    elif len(candidates) < len(existings):
        # more existing than candidate, can't compare (maybe access to
        # previous records has been loast
        yield "u", None, None
    elif len(candidates) == len(existings):
        # same number, we should be able to match them against
        # one another
        yield from _manage_existing(candidates, existings)
    else:
        # at least one existing, but more candidates than existing
        # not supported yet, need to implement incremental intra day
        # push
        logger.debug(
            "got %d candidates for %d existing:\n%r\n%r",
            len(candidates),
            len(existings),
            "\n".join(map(repr, candidates)),
            "\n".join(map(repr, existings)),
        )
        yield "!", None, None


def _get_ff_filters(transactions: TRANSACTIONS):
    dates = []
    for transaction in transactions:
        dates.append(transaction.interest_date)
        dates.append(transaction.date)
    return {"start": min(dates).isoformat(), "end": max(dates).isoformat()}


def get_transactions_to_push(transactions: TRANSACTIONS) -> Commands:
    logger.debug("fetching existing records")
    index: indexes.FireflyIndex = defaultdict(
        lambda: defaultdict(lambda: defaultdict(list))
    )
    ext_id_index: indexes.ExtIdIndex = {}
    transaction_index: indexes.TransactionIndex = defaultdict(
        lambda: defaultdict(lambda: defaultdict(list))
    )

    indexes.feed_tr_indexes(transactions, transaction_index, ext_id_index)

    existing_count = 0
    for page in paginate(
        "get", "transactions", params=_get_ff_filters(transactions)
    ):
        existing_count += 1
        date_key, amount_key, vector_key = indexes.get_ff_keys(page)
        if ext_id_trs := indexes.get_tr_by_ext_ids(ext_id_index, page):
            yield from iter_operations([page], ext_id_trs)
            indexes.clean_tr_indexes(ext_id_trs, transaction_index)
        else:
            index[date_key][amount_key][vector_key].append(page)

    logger.info("got %d exiting records", existing_count)
    for current_date in sorted(transaction_index):
        for amount, trans_by_vector in transaction_index[current_date].items():
            amount = amount * -1 if amount < ZERO else amount
            current_index = index[current_date][amount]
            no_source: Dict[str, TRANSACTIONS] = defaultdict(list)
            no_destination: Dict[str, TRANSACTIONS] = defaultdict(list)
            processed_keys = set()
            for (src_id, dst_id), candidates in trans_by_vector.items():
                if dst_id and src_id is None:
                    no_source[dst_id].extend(candidates)
                elif src_id and dst_id is None:
                    no_destination[src_id].extend(candidates)
                else:
                    processed_keys.add((src_id, dst_id))
                    yield from iter_operations(
                        current_index[(src_id, dst_id)], candidates
                    )
            existings: List[Dict]
            for dst_id, candidates in no_source.items():
                existings = sum(
                    [
                        existing
                        for (
                            esrc_id,
                            edst_id,
                        ), existing in current_index.items()
                        if edst_id == dst_id
                        and (esrc_id, edst_id) not in processed_keys
                    ],
                    [],
                )
                yield from iter_operations(existings, candidates)
            for src_id, candidates in no_destination.items():
                existings = sum(
                    [
                        existing
                        for (
                            esrc_id,
                            edst_id,
                        ), existing in current_index.items()
                        if esrc_id == src_id
                        and (esrc_id, edst_id) not in processed_keys
                    ],
                    [],
                )
                yield from iter_operations(existings, candidates)


def log_body(verb, body, firefly_id=None):
    message = f"{verb} <Tr"
    if firefly_id:
        message += f" {firefly_id}"
    if body.get("group_title"):
        message += f" description={body['group_title']!r}"
    if body.get("transactions"):
        for key, new_value in body["transactions"][0].items():
            message += f" {key}={new_value!r}"
    message += ">"
    logger.warning(message)


def push(transactions: TRANSACTIONS, dry_run: bool = False):
    to_push = list(get_transactions_to_push(transactions))
    if any(operation == "!" for operation, _, _ in to_push):
        logger.fatal("unruly records, aborting!")
        return
    created = updated = 0
    for operation, firefly_id, body in to_push:
        if operation == "create":
            log_body("creating", body)
            if not dry_run:
                query("post", "transactions", body=body)
                created += 1
        elif operation == "update":
            log_body("updating", body, firefly_id)
            if not dry_run:
                query("put", f"transactions/{firefly_id}", body=body)
                updated += 1
    logger.info("created %d new records, updated %d", created, updated)



import datetime
import logging
from dataclasses import asdict, dataclass, field
from decimal import Decimal
from typing import List, Optional, Set

from moula.bootstrap import conf

logger = logging.getLogger(__name__)
ZERO = Decimal("0")
_LOCAL_CACHE: dict = {}


def is_liability(account_id=None, account=None):
    if account_id:
        return _LOCAL_CACHE[f"is-liabilities-{account_id}"]
    liability = account["attributes"]["type"] == "liabilities"
    _LOCAL_CACHE[f"is-liabilities-{account['id']}"] = liability


def set_default_cash_account(account):
    if conf.firefly.default_cash_account:
        return
    if account["attributes"]["type"] == "cash":
        logger.warning("setting default cash account to %r", account["id"])
        _LOCAL_CACHE["id"] = account["id"]


def get_default_cash_account():
    return conf.firefly.default_cash_account or _LOCAL_CACHE.get("id")


@dataclass
class Transaction:
    amount: Decimal
    date: datetime.date
    interest_date: datetime.date
    description: str
    remote_type: str
    currency_code: str

    _external_ids: Set[str] = field(default_factory=set)
    already_merged: bool = False

    source_id: Optional[str] = None
    destination_id: Optional[str] = None
    tags: Optional[List[str]] = None
    payment_date: Optional[datetime.date] = None

    @property
    def payload(self) -> dict:
        exclude_keys = {"_external_ids", "remote_type", "already_merged"}
        properties_to_include = {"type", "external_id"}
        payload = {
            key: value
            for key, value in asdict(self).items()
            if key not in exclude_keys and value
        }
        for key in properties_to_include:
            if getattr(self, key):
                payload[key] = getattr(self, key)
        for key, value in payload.items():
            if key in {"date", "payment_date", "interest_date"}:
                payload[key] = payload[key].isoformat()
        payload["amount"] = str(
            self.amount * (-1 if self.amount < ZERO else 1)
        )
        if payload.get("destination_id") is None:
            payload["destination_id"] = get_default_cash_account()
        if payload.get("source_id") is None:
            payload["source_id"] = get_default_cash_account()
        return payload

    def set_account_id(self, account_id):
        if self.amount < ZERO:
            self.source_id = account_id
        else:
            self.destination_id = account_id

    @property
    def external_id(self):
        return ".".join(sorted(ei for ei in self._external_ids if ei))

    @external_id.setter
    def external_id(self, value: str):
        for sub_value in value.split("."):
            self._external_ids.add(sub_value)

    @property
    def external_ids(self):
        return self._external_ids

    @property
    def type(self):
        if self.source_id and self.destination_id:
            return "transfer"
        if self.amount < ZERO:
            return "withdrawal"
        return "deposit"

    @property
    def is_mergeable(self) -> bool:
        return not self.already_merged

    def __repr__(self):
        msg = f"<Tr dt={self.date} {self.amount}€ "
        if self.source_id:
            msg += f"from={self.source_id} "
        if self.destination_id:
            msg += f"to={self.destination_id} "
        return msg + f"{self.description!r}>"

    def short_print(self, end="\n"):
        print(repr(self), end=end)


TRANSACTIONS = List[Transaction]



from datetime import date
from decimal import Decimal
from moula.bootstrap import conf

ZERO = Decimal("0")
ACCURACY = Decimal(f".{'0' * (conf.accuracy - 1)}1")


def quantize(value: str) -> Decimal:
    return Decimal(value).quantize(ACCURACY)


def remove_double_space(words: str) -> str:
    while "  " in words:
        words = words.replace("  ", " ")
    return words


def french_date_to_iso(string: str, delimiter: str = "/") -> date:
    splits = string.split("/")
    return date(year=int(splits[2]), month=int(splits[1]), day=int(splits[0]))



from decimal import Decimal
from typing import Dict, Optional, Tuple

from moula.types import TRANSACTIONS, Transaction
from moula.utils import ZERO, quantize

VectorKey = Tuple[Optional[str], Optional[str]]
IndexKey = Tuple[str, Decimal, VectorKey]
FireflyIndex = Dict[str, Dict[Decimal, Dict[VectorKey, list]]]
TransactionIndex = Dict[str, Dict[Decimal, Dict[VectorKey, list]]]
ExtIdIndex = Dict[str, Transaction]


def get_ff_keys(page) -> IndexKey:
    trs = page["attributes"]["transactions"]
    amount_key = sum((quantize(tr["amount"]) for tr in trs), ZERO)
    vector_key = (trs[0]["source_id"], trs[0]["destination_id"])
    return trs[0]["date"].split("T")[0], amount_key, vector_key


def get_tr_keys(tr: Transaction) -> IndexKey:
    date_key = getattr(tr, "date").isoformat()
    return date_key, tr.amount, (tr.source_id, tr.destination_id)


def feed_tr_indexes(
    transactions: TRANSACTIONS,
    transaction_index: TransactionIndex,
    ext_id_index: ExtIdIndex,
):
    for tr in transactions:
        date_key, amount_key, vector_key = get_tr_keys(tr)
        transaction_index[date_key][amount_key][vector_key].append(tr)
        for ext_id in tr.external_ids:
            ext_id_index[ext_id] = tr


def clean_tr_indexes(transactions: TRANSACTIONS, index: TransactionIndex):
    for tr in transactions:
        date_key, amount_key, vector_key = get_tr_keys(tr)
        tr_index = index[date_key][amount_key][vector_key]
        if tr in tr_index:
            tr_index.remove(tr)


def get_tr_by_ext_ids(ext_id_index: ExtIdIndex, page) -> TRANSACTIONS:
    ext_id_transactions: TRANSACTIONS = []
    for transaction in page["attributes"]["transactions"]:
        for ext_id in (transaction.get("external_id") or "").split("."):
            match = ext_id_index.get(ext_id) if ext_id else None
            if match and match not in ext_id_transactions:
                ext_id_transactions.append(ext_id_index[ext_id])
    return ext_id_transactions
